#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/types.h>
#include <memory.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>

#include <arch.h>
#include <device.h>

/* Externally used variables: */
byte *HW_OffscreenBuffer=0;
signed_32bit HW_ScreenSizeX;
signed_32bit HW_ScreenSizeY;
signed_32bit HW_ScreenMaxX;
signed_32bit HW_ScreenMaxY;
signed_32bit HW_ScreenCenterX;
signed_32bit HW_ScreenCenterY;
unsigned_32bit HW_BytesPerRow;

/* Internal variables: */
Display		        *disp;
static int              screen;
Window			win;
Colormap		colmap;
GC			cgc;
XShmSegmentInfo         shminfo;
XImage                  *image;
static Visual           *visual;
static long event_mask;

int HW_OpenScreen(void)
{
  XSetWindowAttributes	win_attr;
  XVisualInfo           visual_info;

  HW_ScreenSizeX=640;
  HW_ScreenSizeY=400;
  HW_ScreenMaxX=HW_ScreenSizeX-1;
  HW_ScreenMaxY=HW_ScreenSizeY-1;
  HW_ScreenCenterX=(HW_ScreenSizeX/2)<<4;
  HW_ScreenCenterY=(HW_ScreenSizeY/2)<<4;

  if(!(disp = XOpenDisplay(NULL))) {
    fprintf(stderr, "Can't open X display!\n");
    return 0;
  }
  screen = DefaultScreen (disp);
  win = XCreateSimpleWindow(disp, XDefaultRootWindow(disp), 0, 0,
			    HW_ScreenSizeX, HW_ScreenSizeY, 2, 0, 1);  
  cgc = XCreateGC(disp, win, 0l, NULL);
  win_attr.backing_store = Always;
  XChangeWindowAttributes(disp, win, CWBackingStore , &win_attr);
  XStoreName(disp, win, "Really Coo0l 3-D stuff!");
  XClearWindow(disp, win);
  XMapRaised(disp, win);

  if (!XShmQueryExtension(disp)){
    fprintf(stderr, "No support for shared memory!\n");
    return 0;
  }
  
  if (XMatchVisualInfo (disp, screen, 8 , PseudoColor, &visual_info) == 0){
    fprintf(stderr, "Unable to get 8bit pseudo color!\n");
    return 0;
  }
 
  visual = visual_info.visual;
  image=XShmCreateImage(disp,visual,8,ZPixmap,NULL,&shminfo,HW_ScreenSizeX,HW_ScreenSizeY);
  HW_BytesPerRow=image->bytes_per_line;
  shminfo.shmid=shmget(IPC_PRIVATE,image->bytes_per_line*image->height,
		       IPC_CREAT|0777);
  if (shminfo.shmid < 0) {
    fprintf(stderr, "Unable to get shared memory\n");
    return 0;
  }
  shminfo.shmaddr=image->data=shmat(shminfo.shmid, 0, 0);
  if (shminfo.shmaddr == ((char *) -1)) {
    fprintf(stderr, "Unable to smhat()\n");
    return 0;
  }
  shminfo.readOnly = 0;
  if (XShmAttach(disp,&shminfo)==0) {
    fprintf(stderr, "Unable to attach shared memory\n");
    return 0;
  }
  XFlush(disp); 

  HW_OffscreenBuffer=shminfo.shmaddr;
  return 1;
}

void HW_DisplayScreen(void)
{
  XSetForeground(disp, cgc, 120);

  XShmPutImage(disp,win,cgc,image,0,0,0,0,HW_ScreenSizeX,HW_ScreenSizeY,0); 
  /* Don't generate finished event !!! change this !!!!1*/ 
  XSync(disp, False);
}

void HW_SetPalette(byte *orgpal) 
{
  XSetWindowAttributes	win_attr;
  int i;
  XColor xcol;

  colmap = XCreateColormap(disp, win, DefaultVisual(disp, DefaultScreen(disp)),
			   AllocAll);
  xcol.flags = DoRed | DoGreen | DoBlue;
  for(i = 0; i < 256; i++) {
    xcol.pixel = i;
    xcol.red=  ((int)orgpal[3*i])<<8;
    xcol.green=((int)orgpal[3*i+1])<<8;
    xcol.blue= ((int)orgpal[3*i+2])<<8;
    XStoreColor(disp, colmap, &xcol);
  }

  win_attr.colormap = colmap;
  XChangeWindowAttributes(disp, win, CWColormap, &win_attr);
}

void HW_ClearOffscreen(void)
{
  HW_MemSet(HW_OffscreenBuffer, 0, HW_BytesPerRow*HW_ScreenSizeY);
}

void HW_CloseScreen(void)
{
  XFlush(disp);
  XShmDetach(disp,&shminfo);
  XDestroyImage(image);
  shmdt(shminfo.shmaddr);
  shmctl(shminfo.shmid,IPC_RMID,0);
  XFreeColormap(disp, colmap);
  XFreeGC(disp, cgc); /* forgets to free gc[] */
  XDestroyWindow(disp, win);
  XCloseDisplay(disp);
}

