/*

TODO:
  * VCross()
*/
#include <math.h>

#include <arch.h>
#include <error.h>
#include <dbgmem.h>

#include "vector.h"

/* Global Variables: */

/* vectors and matrix functions */
void MCopy(matrix source,matrix dest)
{
  memcpy(dest,source,sizeof(matrix));
}

int MCompare(matrix a, matrix b)
{
  int i,j;

  for (i=0;i<3;i++) {
    for (j=0;j<4;j++) {
      if (a[i][j]!=b[i][j])
	return 0;
    }
  }
  return 1;
}

int VCompare(vector *a, vector *b)
{
  if (a->x!=b->x)
    return 0;
  if (a->y!=b->y)
    return 0;
  if (a->z!=b->z)
    return 0;
  return 1;
}

void MMult(matrix a,matrix b,matrix c)
{
#ifdef i386
  register float ka1,kb1,kc1;
  
  ka1=a[0][0];kb1=a[0][1];kc1=a[0][2];
  c[0][0]= ka1*b[0][0] + kb1*b[1][0] + kc1*b[2][0];
  c[0][1]= ka1*b[0][1] + kb1*b[1][1] + kc1*b[2][1];
  c[0][2]= ka1*b[0][2] + kb1*b[1][2] + kc1*b[2][2];
  c[0][3]= ka1*b[0][3] + kb1*b[1][3] + kc1*b[2][3] + a[0][3];
  
  c[1][0]= a[1][0]*b[0][0] + a[1][1]*b[1][0] + a[1][2]*b[2][0];
  c[1][1]= a[1][0]*b[0][1] + a[1][1]*b[1][1] + a[1][2]*b[2][1];
  c[1][2]= a[1][0]*b[0][2] + a[1][1]*b[1][2] + a[1][2]*b[2][2];
  c[1][3]= a[1][0]*b[0][3] + a[1][1]*b[1][3] + a[1][2]*b[2][3] + a[1][3];
  
  c[2][0]= a[2][0]*b[0][0] + a[2][1]*b[1][0] + a[2][2]*b[2][0];
  c[2][1]= a[2][0]*b[0][1] + a[2][1]*b[1][1] + a[2][2]*b[2][1];
  c[2][2]= a[2][0]*b[0][2] + a[2][1]*b[1][2] + a[2][2]*b[2][2];
  c[2][3]= a[2][0]*b[0][3] + a[2][1]*b[1][3] + a[2][2]*b[2][3] + a[2][3];

#else
  {
    register float ka1,kb1,kc1;
    
    ka1=a[0][0];kb1=a[0][1];kc1=a[0][2];
    c[0][0]= ka1*b[0][0] + kb1*b[1][0] + kc1*b[2][0];
    c[0][1]= ka1*b[0][1] + kb1*b[1][1] + kc1*b[2][1];
    c[0][2]= ka1*b[0][2] + kb1*b[1][2] + kc1*b[2][2];
    c[0][3]= ka1*b[0][3] + kb1*b[1][3] + kc1*b[2][3] + a[0][3];
  }

  {
    register float ka2,kb2,kc2;
    ka2=a[1][0];kb2=a[1][1];kc2=a[1][2];
    c[1][0]= ka2*b[0][0] + kb2*b[1][0] + kc2*b[2][0];
    c[1][1]= ka2*b[0][1] + kb2*b[1][1] + kc2*b[2][1];
    c[1][2]= ka2*b[0][2] + kb2*b[1][2] + kc2*b[2][2];
    c[1][3]= ka2*b[0][3] + kb2*b[1][3] + kc2*b[2][3] + a[1][3];
  }
  {
    register float ka3,kb3,kc3;
    ka3=a[2][0];kb3=a[2][1];kc3=a[2][2];
    c[2][0]= ka3*b[0][0] + kb3*b[1][0] + kc3*b[2][0];
    c[2][1]= ka3*b[0][1] + kb3*b[1][1] + kc3*b[2][1];
    c[2][2]= ka3*b[0][2] + kb3*b[1][2] + kc3*b[2][2];
    c[2][3]= ka3*b[0][3] + kb3*b[1][3] + kc3*b[2][3] + a[2][3];
  }
#endif
}

void VMult4(matrix a,vector *b,vector *c)
{
  c->x= a[0][0]*b->x + a[0][1]*b->y + a[0][2]*b->z + a[0][3];
  c->y= a[1][0]*b->x + a[1][1]*b->y + a[1][2]*b->z + a[1][3];
  c->z= a[2][0]*b->x + a[2][1]*b->y + a[2][2]*b->z + a[2][3];
}

void VMult3(matrix a,vector *b,vector *c)
{
  c->x= a[0][0]*b->x + a[0][1]*b->y + a[0][2]*b->z;
  c->y= a[1][0]*b->x + a[1][1]*b->y + a[1][2]*b->z;
  c->z= a[2][0]*b->x + a[2][1]*b->y + a[2][2]*b->z;
}

void MRotX(real theta,matrix dest)
{
  dest[0][0]=1.0;
  dest[0][1]=0;
  dest[0][2]=0;
  dest[0][3]=0;

  dest[1][0]=0;
  dest[1][1]=cos(theta);
  dest[1][2]=-sin(theta);
  dest[1][3]=0;

  dest[2][0]=0;
  dest[2][1]=sin(theta);
  dest[2][2]=cos(theta);
  dest[2][3]=0;
}

void MRotY(real theta,matrix dest)
{
  dest[0][0]=cos(theta);
  dest[0][1]=0;
  dest[0][2]=sin(theta);
  dest[0][3]=0;

  dest[1][0]=0;
  dest[1][1]=1.0;
  dest[1][2]=0;
  dest[1][3]=0;

  dest[2][0]=-sin(theta);
  dest[2][1]=0;
  dest[2][2]=cos(theta);
  dest[2][3]=0;
}

void MRotZ(real theta,matrix dest)
{
  dest[0][0]=cos(theta);
  dest[0][1]=-sin(theta);
  dest[0][2]=0;
  dest[0][3]=0;

  dest[1][0]=sin(theta);
  dest[1][1]=cos(theta);
  dest[1][2]=0;
  dest[1][3]=0;

  dest[2][0]=0;
  dest[2][1]=0;
  dest[2][2]=1.0;
  dest[2][3]=0;
}

real VDot(vector *a,vector *b)
{
  return a->x*b->x + a->y*b->y + a->z*b->z;
}

void VCross(vector *a,vector *b,vector *c)
{
  assert(0);
}

void MOrtoNormalize(matrix a)
{
  real norm,proj;
  
/* normalize first vector: */
  norm=a[0][0]*a[0][0] + a[1][0]*a[1][0] + a[2][0]*a[2][0];
  norm=sqrt(norm);
  a[0][0]=a[0][0]/norm;
  a[1][0]=a[1][0]/norm;
  a[2][0]=a[2][0]/norm;
  
/* project second vector on first */
  proj=a[0][1]*a[0][0] + a[1][1]*a[1][0] + a[2][1]*a[2][0];
/* Remove projected part */
  a[0][1]-=proj*a[0][0];
  a[1][1]-=proj*a[1][0];
  a[2][1]-=proj*a[2][0];

/* normalize second vector: */
  norm=a[0][1]*a[0][1] + a[1][1]*a[1][1] + a[2][1]*a[2][1];
  norm=sqrt(norm);
  a[0][1]=a[0][1]/norm;
  a[1][1]=a[1][1]/norm;
  a[2][1]=a[2][1]/norm;

/* Use Cross product to calculate last vector */
  a[0][2]=a[1][0]*a[2][1] - a[2][0]*a[1][1];
  a[1][2]=a[2][0]*a[0][1] - a[0][0]*a[2][1];
  a[2][2]=a[0][0]*a[1][1] - a[1][0]*a[0][1];
}

real DetM(matrix a)
{
  return
    a[0][0]*(a[1][1]*a[2][2] - a[1][2]*a[2][1]) - 
    a[0][1]*(a[1][0]*a[2][2] - a[2][0]*a[1][2]) +
    a[0][2]*(a[1][0]*a[2][1] - a[2][0]*a[1][1]);
}

void InvertM(matrix a)
{
  real d;
  matrix b,c;
  vector v,u;

  b[0][0]= (a[1][1]*a[2][2]) - (a[1][2]*a[2][1]);
  b[0][1]= (a[2][0]*a[1][2]) - (a[1][0]*a[2][2]);
  b[0][2]= (a[1][0]*a[2][1]) - (a[2][0]*a[1][1]);
  b[1][0]=-(a[0][1]*a[2][2]) + (a[2][1]*a[0][2]);
  b[1][1]= (a[0][0]*a[2][2]) - (a[2][0]*a[0][2]);
  b[1][2]=-(a[0][0]*a[2][1]) + (a[2][0]*a[0][1]);
  b[2][0]= (a[0][1]*a[1][2]) - (a[1][1]*a[0][2]);
  b[2][1]=-(a[0][0]*a[1][2]) + (a[1][0]*a[0][2]);
  b[2][2]= (a[0][0]*a[1][1]) - (a[1][0]*a[0][1]);


  d=a[0][0]*b[0][0] + a[0][1]*b[0][1] + a[0][2]*b[0][2];

  assert(d!=0);
  d=1.0/d;

  c[0][0]=d*b[0][0];
  c[0][1]=d*b[1][0];
  c[0][2]=d*b[2][0];
  c[1][0]=d*b[0][1];
  c[1][1]=d*b[1][1];
  c[1][2]=d*b[2][1];
  c[2][0]=d*b[0][2];
  c[2][1]=d*b[1][2];
  c[2][2]=d*b[2][2];

  v.x=a[0][3];
  v.y=a[1][3];
  v.z=a[2][3];

  VMult3(c,&v,&u);

  c[0][3]=-u.x;
  c[1][3]=-u.y;
  c[2][3]=-u.z;

  MCopy(c,a);
}
